---- Customized scoring

require("glon")

SCORE = {}
SCORE.Events = {}
--SCORE.Scores = {}

-- One might wonder why all the key names in the event tables are so annoyingly
-- short. Well, the serialisation module in gmod (glon) does not do any
-- compression. At all. This means the difference between all events having a
-- "time_added" key versus a "t" key is very significant for the amount of data
-- we need to send. It's a pain, but I'm not going to code my own compression,
-- so doing it manually is the only way.

-- One great way to reduce data sent turned out to be rounding the time floats.
-- We don't actually need to know about 10000ths of seconds after all.

-- From gmod wiki
function math.AdvRound( val, d )
   d = d or 0;
   return math.Round( val * (10 ^ d) ) / (10 ^ d);
end

function SCORE:AddEvent(entry)
   entry["t"] = math.AdvRound(CurTime(), 2)
   table.insert(self.Events, entry)
end

local function CopyDmg(dmg)

   local wep = util.WeaponFromDamage(dmg)

   -- t = type, a = amount, g = gun, h = headshot
   local d = {}
--   d.t = (dmg:IsFallDamage() and DMG_FALL) or (dmg:IsBulletDamage() and DMG_BULLET) or (dmg:IsExplosionDamage() and DMG_BLAST)
   d.t = dmg:GetDamageType()
   d.a = dmg:GetDamage()
   d.h = false

   if wep then
      d.g = WepToEnum(wep)
   else
      d.g = -1
   end

   return d
end

function SCORE:HandleKill( victim, attacker, dmginfo )

   local e = {
      id=EVENT_KILL, 
      att={ni="", uid=-1, tr=false}, 
      vic={ni=victim:Nick(), uid=victim:UniqueID(), tr=false}, 
      dmg=CopyDmg(dmginfo)};

   if ValidEntity(attacker) and attacker:IsPlayer() then
      e.att.ni = attacker:Nick()
      e.att.uid = attacker:UniqueID()
      e.att.tr = attacker:GetTraitor()
   end

   if victim:IsPlayer() then
--      Msg("Kill headshot val: " .. tostring(victim.was_headshot) .. "\n")
      e.dmg.h = victim.was_headshot
   end

   e.vic.tr = victim:GetTraitor()

   self:AddEvent(e)
end

function SCORE:HandleSpawn( ply )
   if ply:Team() == TEAM_TERROR then
      self:AddEvent({id=EVENT_SPAWN, ni=ply:Nick(), uid=ply:UniqueID()})
   end
end

function SCORE:HandleTraitorSelection()
   local traitors = {}
   for k, ply in pairs(player.GetAll()) do
      if ply.is_traitor then
         table.insert(traitors, ply:UniqueID())
      end
   end

   self:AddEvent({id=EVENT_SELECTED, traitor_ids=traitors})
end

function SCORE:HandleBodyFound(finder, found)
   self:AddEvent({id=EVENT_BODYFOUND, ni=finder:Nick(), uid=finder:UniqueID(), b=found:Nick()})
end

function SCORE:ApplyEventLogScores()
   local scores = {}
   local traitors = {}
   for k, ply in pairs(player.GetAll()) do
      scores[ply:UniqueID()] = {}

      if ply:GetTraitor() then
         table.insert(traitors, ply:UniqueID())
      end
   end

   -- individual scores, and count those left alive
   local alive = {traitors = 0, innos = 0}
   local dead = {traitors = 0, innos = 0}
   local scored_log = ScoreEventLog(self.Events, scores, traitors)
   local ply = nil
   for id, s in pairs(scored_log) do
      ply = player.GetByUniqueID(id)
      if ply and ply:ShouldScore() then
         ply:AddFrags(KillsToPoints(s, ply:GetTraitor()))
      end
   end

   -- team scores
   local bonus = ScoreTeamBonus(scored_log)

   for id, s in pairs(scored_log) do
      ply = player.GetByUniqueID(id)
      if ply and ply:ShouldScore() then
         ply:AddFrags(ply:GetTraitor() and bonus.traitors or bonus.innos)
      end
   end

   -- count deaths
   for k, e in pairs(self.Events) do
      if e.id == EVENT_KILL then
         local victim = player.GetByUniqueID(e.vic.uid)
         if ValidEntity(victim) then
            victim:AddDeaths(1)
         end
      end
   end
end

function SCORE:RoundStateChange(newstate)
   self:AddEvent({id=EVENT_GAME, state=newstate})
end

function SCORE:RoundComplete(wintype)
   self:AddEvent({id=EVENT_FINISH, win=wintype})
end

function SCORE:Reset()
   self.Events = {}
end

local function CompressForStream(events)
   -- may want to filter out data later
   -- just serialize for now

--   local h_a, h_b, h_c = debug.gethook()
--   debug.sethook()
   local status, result = pcall(glon.encode, events)
--   debug.sethook(h_a, h_b, h_c)

   if not status then
      ErrorNoHalt("Round report event encoding returned an error: " .. result .. "\n")
      return false
   else
      return result
   end
end

function SCORE:StreamToClients()
--   datastream.StreamToClients(player.GetHumans(), "ReportStreamHook", self.Events)
   -- datastream is a piece of crap, do it ourselves

   local s = CompressForStream(self.Events)

   if not s then
      return -- error occurred
   end

   -- divide into happy lil bits
   local cut = {}
   while #s != 0 do
      local bit = string.sub(s, 1, 127)
      table.insert(cut, bit)

      s = string.sub(s, 128, -1)
   end
   
   local parts = #cut
--   Msg("Sending msg in parts: " .. parts.. "\n")
   for k, bit in pairs(cut) do
      
      umsg.Start("report_stream")
      umsg.Bool( (k != parts) ) -- continuation bit, true if there's more coming
      umsg.String(bit)
      umsg.End()
   end
   
--   Msg("Message fully sent\n")
end
